/*

Copyright (C) 2012  Observatoire de Paris

The JavaScript code in this page is free software: you can
redistribute it and/or modify it under the terms of the GNU
General Public License (GNU GPL) as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.  The code is distributed WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.

As additional permission under GNU GPL version 3 section 7, you
may distribute non-source (e.g., minimized or compacted) forms of
that code without the copy of the GNU GPL normally required by
section 4, provided you include this license notice and a URL
through which recipients can access the Corresponding Source.

*/

var valeurAffichee; // String
var expressionCachee; // String
var valeur1; // double
var valeur2; // double
var edit; // boolean
var opCourante; // Fonction
var memoire; // double
var modeExpressions; // boolean
var degrad; // String
var constantes; // Array of Constante
var idVersTruc; // Array[String] of Truc
var theTarget, checkTarget;

function derivation(enfant, parent) {
    // la "classe" enfant dérive de la "classe" parent
    var tmp = function () {}
    tmp.prototype = parent.prototype;
    enfant.prototype = new tmp();
    enfant.prototype.constructor = enfant;
}

function init() {
    valeur2 = 0;
    opCourante = null;
    valeur1 = 0;
    valeurAffichee = "0";
    expressionCachee = "";
    edit = false;
    memoire = 0;
    modeExpressions = false;
    degrad = "rad";
    
    trucs = new Array();
    trucs.push(new Addition());
    trucs.push(new Soustraction());
    trucs.push(new Multiplication());
    trucs.push(new Division());
    trucs.push(new Puissance());
    trucs.push(new Inverse());
    trucs.push(new PuissanceInverse());
    trucs.push(new Factorielle());
    trucs.push(new M());
    trucs.push(new RM());
    trucs.push(new Mplus());
    trucs.push(new Pourcent());
    trucs.push(new Expressions());
    trucs.push(new Lettre(','));
    trucs.push(new DMS());
    trucs.push(new SigneInverse());
    trucs.push(new DegRad());
    trucs.push(new Del());
    trucs.push(new C());
    trucs.push(new AC());
    trucs.push(new Sinus());
    trucs.push(new ArcSinus());
    trucs.push(new Cosinus());
    trucs.push(new ArcCosinus());
    trucs.push(new Tangente());
    trucs.push(new ArcTangente());
    trucs.push(new Exponentielle());
    trucs.push(new LogarithmeNeperien());
    trucs.push(new Exp10());
    trucs.push(new LogarithmeDecimal());
    trucs.push(new Carre());
    trucs.push(new RacineCarree());
    trucs.push(new Lettre('('));
    trucs.push(new Lettre(')'));
    trucs.push(new Lettre('0'));
    trucs.push(new Lettre('1'));
    trucs.push(new Lettre('2'));
    trucs.push(new Lettre('3'));
    trucs.push(new Lettre('4'));
    trucs.push(new Lettre('5'));
    trucs.push(new Lettre('6'));
    trucs.push(new Lettre('7'));
    trucs.push(new Lettre('8'));
    trucs.push(new Lettre('9'));
    trucs.push(new Lettre('.'));
    trucs.push(new Lettre('E'));
    trucs.push(new Lettre('='));
    
    idVersTruc = new Array();
    for (var i=0; i<trucs.length; i++)
        idVersTruc[trucs[i].getId()] = trucs[i];
    
    constantes = new Array();
    ajouterConstante("π", "π", Math.PI);
    ajouterConstante("e", "e", Math.E);
    
    ajouterConstante("c", "c (m/s)", 2.99792458E8);
    ajouterConstante("G", "G (kg-1 m3 s-2)", 6.67384E-11); // CODATA 2010 ; IAU 2009 = CODATA 2006 : 6.67428E-11
    ajouterConstante("h", "h (J s)", 6.6260755E-34);
    ajouterConstante("k", "k (J/K)", 1.380622E-23);
    
    ajouterConstante("c_e", "c_e (C)", 1.60217733E-19);
    ajouterConstante("m_e", "m_e (kg)", 9.1093897E-31);
    ajouterConstante("m_p", "m_p (kg)", 1.6726231E-27);
    ajouterConstante("m_n", "m_n (kg)", 1.6749286E-27);
    
    ajouterConstante("UA", "UA (m)",  149597870700); // définition UAI 2012
    ajouterConstante("AL", "AL (m)", 9.463E15);
    ajouterConstante("pc", "pc (m)", 3.0856775807E16);
    
    // données sur les planètes issues de
    // http://portail.imcce.fr/fr/ephemerides/astronomie/Promenade/pages1/19.html
    //MenuConst mplanetes = new MenuConst("système solaire");
    //MenuConst mlune = new MenuConst("Lune");
    ajouterConstante("alun", "Lune - demi-grand axe (m)", 3.83398E8);
    ajouterConstante("elun", "Lune - excentricité", 0.05555);
    ajouterConstante("Tlun", "Lune - période sid. (j)", 27.3217); // période de révolution sidérale (en jours terrestre)
    ajouterConstante("dlun", "Lune - diamètre eq. (m)", 3.4442E6); // diametre équatorial
    ajouterConstante("mlun", "Lune - masse (kg)", 7.34753E22);
    //mplanetes.ajouter(mlune);
    //MenuConst msoleil = new MenuConst("Soleil");
    ajouterConstante("msol", "Soleil - masse (kg)", 1.9887973E30);
    ajouterConstante("dsol", "Soleil - diamètre eq. (m)", 1.392E9);
    ajouterConstante("Tsol", "Soleil - température (K)", 5780); // température de surface
    ajouterConstante("Lsol", "Soleil - luminosité (W)", 3.846E26); // luminosité
    //mplanetes.ajouter(msoleil);
    //MenuConst mmercure = new MenuConst("Mercure");
    ajouterConstante("amer", "Mercure - demi-grand axe (UA)", 0.3870983);
    ajouterConstante("emer", "Mercure - excentricité", 0.20563);
    ajouterConstante("Tmer", "Mercure - période sid. (j)", 87.9693); // période de révolution sidérale (en jours terrestre)
    ajouterConstante("dmer", "Mercure - diamètre eq. (m)", 4.8794E6); // diametre équatorial
    ajouterConstante("mmer", "Mercure - masse (kg)", 3.3018E23); // masse sans compter les satellites
    //mplanetes.ajouter(mmercure);
    //MenuConst mvenus = new MenuConst("Vénus");
    ajouterConstante("aven", "Vénus - demi-grand axe (UA)", 0.7233298);
    ajouterConstante("even", "Vénus - excentricité", 0.00677);
    ajouterConstante("Tven", "Vénus - période sid. (j)", 224.701);
    ajouterConstante("dven", "Vénus - diamètre eq. (m)", 12.1036E6);
    ajouterConstante("mven", "Vénus - masse (kg)", 4.8685E24);
    //mplanetes.ajouter(mvenus);
    //MenuConst mterre = new MenuConst("Terre");
    ajouterConstante("ater", "Terre - demi-grand axe (UA)", 1.0000010);
    ajouterConstante("eter", "Terre - excentricité", 0.01671);
    ajouterConstante("Tter", "Terre - période sid. (j)", 365.256);
    ajouterConstante("dter", "Terre - diamètre eq. (m)", 12.75628E6);
    ajouterConstante("mter", "Terre - masse (kg)", 5.9736E24);
    //mplanetes.ajouter(mterre);
    //MenuConst mmars = new MenuConst("Mars");
    ajouterConstante("amar", "Mars - demi-grand axe (UA)", 1.5236793);
    ajouterConstante("emar", "Mars - excentricité", 0.09340);
    ajouterConstante("Tmar", "Mars - période sid. (j)", 686.986);
    ajouterConstante("dmar", "Mars - diamètre eq. (m)", 6.794E6);
    ajouterConstante("mmar", "Mars - masse (kg)", 6.4185E23);
    //mplanetes.ajouter(mmars);
    //MenuConst mjupiter = new MenuConst("Jupiter");
    ajouterConstante("ajup", "Jupiter - demi-grand axe (UA)", 5.202603);
    ajouterConstante("ejup", "Jupiter - excentricité", 0.04850);
    ajouterConstante("Tjup", "Jupiter - période sid. (j)", 4.332656E3);
    ajouterConstante("djup", "Jupiter - diamètre eq. (m)", 1.42984E8);
    ajouterConstante("mjup", "Jupiter - masse (kg)", 1.8986E27);
    //mplanetes.ajouter(mjupiter);
    //MenuConst msaturne = new MenuConst("Saturne");
    ajouterConstante("asat", "Saturne - demi-grand axe (UA)", 9.554909);
    ajouterConstante("esat", "Saturne - excentricité", 0.05555);
    ajouterConstante("Tsat", "Saturne - période sid. (j)", 1.0759404E4);
    ajouterConstante("dsat", "Saturne - diamètre eq. (m)", 1.20536E8);
    ajouterConstante("msat", "Saturne - masse (kg)", 5.6846E26);
    //mplanetes.ajouter(msaturne);
    //MenuConst muranus = new MenuConst("Uranus");
    ajouterConstante("aura", "Uranus - demi-grand axe (UA)", 19.21845);
    ajouterConstante("eura", "Uranus - excentricité", 0.04638);
    ajouterConstante("Tura", "Uranus - période sid. (j)", 3.0688984E4);
    ajouterConstante("dura", "Uranus - diamètre eq. (m)", 5.1118E7);
    ajouterConstante("mura", "Uranus - masse (kg)", 8.6831E25);
    //mplanetes.ajouter(muranus);
    //MenuConst mneptune = new MenuConst("Neptune");
    ajouterConstante("anep", "Neptune - demi-grand axe (UA)", 30.11039);
    ajouterConstante("enep", "Neptune - excentricité", 0.00946);
    ajouterConstante("Tnep", "Neptune - période sid. (j)", 6.0183284E4);
    ajouterConstante("dnep", "Neptune - diamètre eq. (m)", 4.9528E7);
    ajouterConstante("mnep", "Neptune - masse (kg)", 1.0243E26);
    //mplanetes.ajouter(mneptune);
    //MenuConst mpluton = new MenuConst("Pluton");
    ajouterConstante("aplu", "Pluton - demi-grand axe (UA)", 39.54470);
    ajouterConstante("eplu", "Pluton - excentricité", 0.2490);
    ajouterConstante("Tplu", "Pluton - période sid. (j)", 9.0580232E4);
    ajouterConstante("dplu", "Pluton - diamètre eq. (m)", 2.390E6);
    ajouterConstante("mplu", "Pluton - masse (kg)", 1.238E22);
    //mplanetes.ajouter(mpluton);
    //mconstantes.ajouter(mplanetes);
    
    ajouterConstante("answer", "ultimate answer", 42);
    
    /*
    on ne peut pas changer directement le select avec ie
    cf bug http://support.microsoft.com/kb/276228
    */
    var htmlConstantes = "<select id=\"constantes\" onchange=\"choixConstante(this.selectedIndex)\"><option>constante<\/option>";
    for (var i=0; i<constantes.length; i++) {
        var cst = constantes[i];
        htmlConstantes += "<option>" + cst.getTitre() + "<\/option>";
    }
    htmlConstantes += "<\/select>";
    var divConstantes = document.getElementById('divconstantes');
    divConstantes.innerHTML = htmlConstantes;
    
    if (document.addEventListener) {
        document.addEventListener('keypress', keypress, false);
        document.addEventListener('keyup', keyup, false);
    } else if (document.attachEvent) {
        document.attachEvent('onkeypress', keypress);
        document.attachEvent('onkeyup', keyup);
    }
    evenementsBoutons();
}

function ajouterConstante(nom, titre, valeur) {
    constantes.push(new Constante(nom, titre, valeur));
}

function chercherConstante(nom) {
    for (var i=0; i<constantes.length; i++) {
        if (constantes[i].getNom().toUpperCase() === nom.toUpperCase())
            return(constantes[i]);
    }
    return(null);
}

//public versString(double v) {
function versString(v) {
    var precision = 10;
    var limitesci = 3; // décimales avant de passer en mode sci
    
    if (v == 0)
        return("0");
    var sj = ""+v;
    if ("Infinity" === sj || "-Infinity" === sj || "NaN" === sj)
        return(""+v);
    var signe;
    if (v >= 0)
        signe = "";
    else {
        v = -v;
        signe = "-";
    }
    var exposant = Math.floor(Math.log(v)/Math.log(10));
    var exparr = Math.pow(10, precision);
    var mantisse = v / Math.pow(10, exposant);
    //mantisse = Math.rint(mantisse*exparr)/exparr;
    mantisse = Math.round(mantisse*exparr)/exparr;
    if (mantisse >= 10) {
        exposant += 1;
        mantisse = v / Math.pow(10, exposant);
        //mantisse = Math.rint(mantisse*exparr)/exparr;
        mantisse = Math.round(mantisse*exparr)/exparr;
    }
    var sv = "";
    var chiffre = Math.floor(mantisse);
    sv += chiffre + ".";
    for (var i=0; i<precision-1; i++) {
        mantisse = (mantisse - chiffre)*10;
        var exparr2 = Math.pow(10, precision-i);
        //mantisse = Math.rint(mantisse*exparr)/exparr;
        mantisse = (Math.round(mantisse*exparr))/exparr;
        chiffre = Math.floor(mantisse);
        sv += chiffre;
    }
    mantisse = (mantisse - chiffre)*10;
    chiffre = Math.round(mantisse);
    sv += chiffre;
    var dernier = sv.charAt(sv.length-1);
    while (dernier == '0') {
        sv = sv.substring(0, sv.length-1);
        dernier = sv.charAt(sv.length-1);
    }
    if (exposant != 0) {
        if (exposant > 0 && exposant < limitesci) {
            var sd = sv.substring(0,1);
            if (sv.length-2 < exposant) {
                // transformation de 3.2E2 en 320
                if (sv.length > 2)
                    sd += sv.substring(2,sv.length);
                for (var i=0; i<exposant-(sv.length-2); i++)
                    sd += '0';
            } else {
                // transformation de 3.245E2 en 324.5
                if (sv.length < exposant+2)
                    sd += sv.substring(2);
                else
                    sd += sv.substring(2,exposant+2) + "." + sv.substring(exposant+2);
            }
            sv = sd;
        } else if (exposant < 0 && exposant > -limitesci) {
            // transformation de 3.2E-2 en 0.032
            var sd = "0.";
            for (var i=0; i<-exposant-1; i++)
                sd += '0';
            sv = sd + sv.charAt(0) + sv.substring(2);
        } else
            sv += "E" + exposant;
    }
    if (sv.charAt(sv.length-1) == '.')
        sv = sv.substring(0, sv.length-1);
    
    sv = signe + sv;
    
    return(sv);
}

//public void operation(Truc truc) {
function operation(truc) {
    if (truc instanceof Lettre && "=" === truc.getNom()) {
        if (modeExpressions) {
            var valeur = calculer(valeurAffichee);
            if (valeur == null)
                valeurAffichee = "erreur";
            else {
                //valeur1 = valeur.doubleValue();
                valeur1 = valeur;
                valeurAffichee = versString(valeur1);
                expressionCachee = "" + valeur1;
            }
            valeur2 = 0;
            opCourante = null;
            edit = false;
        } else {
            //try {
                if (edit) {
                    //valeur1 = (new Double(valeurAffichee)).doubleValue();
                    valeur1 = Number(valeurAffichee);
                    expressionCachee += valeur1;
                    if (opCourante != null && !opCourante.estOperateur())
                        expressionCachee += ')';
                }
                if (opCourante != null) {
                    var valeur = calculer(expressionCachee);
                    if (valeur == null)
                        valeurAffichee = "erreur";
                    else {
                        //valeur1 = valeur.doubleValue();
                        valeur1 = valeur;
                        valeurAffichee = versString(valeur1);
                    }
                }
                expressionCachee = "" + valeur1;
                valeur2 = 0;
                opCourante = null;
                edit = false;
            //} catch (ex) {
            //    alert("operation1: NumberFormatException: " + ex.description);
            //}
        }
    } else if (truc instanceof Lettre) {
        var nom = truc.getNom();
        if (!modeExpressions && ")" === nom && edit) {
            try {
                //valeur1 = (new Double(valeurAffichee)).doubleValue();
                valeur1 = Number(valeurAffichee);
                expressionCachee += valeur1;
            } catch (ex) {
                alert("operation2: NumberFormatException: " + ex.description);
            }
        }
        if (!edit) {
            valeurAffichee = "";
            if (opCourante == null)
                expressionCachee = "";
            edit = true;
        }
        if (!modeExpressions && ("(" === nom || ")" === nom)) {
            if (")" === nom) {
                valeurAffichee = "";
                edit = false;
            }
            expressionCachee += nom;
        }
        if (modeExpressions || ("(" !== nom && ")" !== nom))
            valeurAffichee += nom;
    } else if (truc instanceof Special) {
        truc.executer();
    } else if (truc instanceof Fonction) {
        var fct = truc;
        if (modeExpressions) {
            if (!edit && !fct.estOperateur())
                valeurAffichee = "";
            valeurAffichee += fct.getNom();
            if (!fct.estOperateur())
                valeurAffichee += "(";
            edit = true;
        } else {
            // cas spécial du - après un E
            if (("-" === fct.getNom() || "signeInverse" === fct.getNom()) &&
                    (valeurAffichee.lastIndexOf("E") == valeurAffichee.length-1) && edit)
                valeurAffichee += "-";
            else {
                try {
                    if (edit) {
                        try {
                            //valeur1 = (new Double(valeurAffichee)).doubleValue();
                            valeur1 = Number(valeurAffichee);
                        } catch (ex) {
                            alert("operation3: NumberFormatException: " + ex.description);
                        }
                    }
                    if (fct.getNbParams() == 0) {
                        valeur1 = fct.evaluer(0, 0);
                        valeurAffichee = versString(valeur1);
                        if (!edit && opCourante == null)
                            expressionCachee = "";
                        edit = true;
                    } else if (fct.getNbParams() == 1) {
                        valeur1 = fct.evaluer(valeur1, 0);
                        valeurAffichee = versString(valeur1);
                        if (!edit)
                            expressionCachee = "" + valeur1;
                    } else {
                        if (fct.estOperateur()) {
                            if (edit)
                                expressionCachee += valeur1 + fct.getNom();
                            else
                                expressionCachee += fct.getNom();
                        } else {
                            if (edit)
                                expressionCachee += fct.getNom() + '(' + valeur1 + ',';
                            else
                                expressionCachee = fct.getNom() + '(' + expressionCachee + ',';
                        }
                        valeur2 = valeur1;
                        opCourante = fct;
                        edit = false;
                    }
                } catch (ex) {
                    alert("operation4: NumberFormatException: " + ex.description);
                }
            }
        }
    }
    affichage();
}

function affichage() {
    var valeurspan = document.getElementById('valeur');
    if (valeurAffichee.length < 30)
        valeurspan.innerHTML = valeurAffichee;
    else
        valeurspan.innerHTML = "..." + valeurAffichee.substring(valeurAffichee.length - 29, valeurAffichee.length);
    var infos = degrad;
    if (modeExpressions)
        infos += " expr";
    var infospan = document.getElementById('infos');
    infospan.innerHTML = infos;
    var memspan = document.getElementById('memoire');
    if (memoire != 0)
        memspan.innerHTML = "M";
    else
        memspan.innerHTML = "&nbsp;";
}

//public Truc chercherTruc(String nom) {
function chercherTruc(nom) {
    for (var i=0; i<trucs.length; i++) {
        if (trucs[i].getNom().toUpperCase() === nom.toUpperCase())
            return(trucs[i]);
    }
    return(null);
}


//abstract class Truc {
function Truc(id, nom, titre) {
    this.id = id;
    this.nom = nom;
    this.titre = titre;
    this.actif = false;
    return(this);
}
Truc.prototype.getId = function() {
    return(this.id);
}
Truc.prototype.getNom = function() {
    return(this.nom);
}
Truc.prototype.getTitre = function() {
    return(this.titre);
};
Truc.prototype.getActif = function() {
    return(this.actif);
};
Truc.prototype.activation = function(activer) {
    if (this.id != null) {
        var bouton = document.getElementById(this.id);
        if (bouton != null) {
            if (activer) {
                if (!this.actif) {
                    this.actif = true;
                    bouton.className += " actif";
                }
            } else {
                this.actif = false;
                bouton.className = bouton.className.replace(" actif", "");
            }
        }
    }
}


//abstract class Fonction extends Truc {
derivation(Fonction, Truc);
function Fonction(id, nom, titre, operateur, nbparams) {
    Truc.call(this, id, nom, titre);
    this.operateur = operateur;
    this.nbparams = nbparams;
    return(this);
}
Fonction.prototype.estOperateur = function() {
    return(this.operateur);
}
Fonction.prototype.getNbParams = function() {
    return(this.nbparams);
}
//public abstract double evaluer(double v1, double v2);


//abstract class Special extends Truc {
derivation(Special, Truc);
function Special(id, nom, titre) {
    Truc.call(this, id, nom, titre);
    return(this);
}
//public abstract void executer();


//public Fonction obtenirFonction(String nom) {
function obtenirFonction(nom) {
    var truc = chercherTruc(nom);
    if (truc == null) {
        alert("opération non trouvée: " + nom);
        return(null);
    }
    if (!(truc instanceof Fonction)) {
        alert("opération de mauvais type: " + nom);
        return(null);
    }
    return(truc);
}


//class Lettre extends Truc {
derivation(Lettre, Truc);
function Lettre(c) {
    var id;
    if (c == "0" || c == "1" || c == "2" || c == "3" || c == "4" ||
            c == "5" || c == "6" || c == "7" || c == "8" || c == "9")
        id = "chiffre" + c;
    else if (c == "(")
        id = "pargauche";
    else if (c == ")")
        id = "pardroite";
    else if (c == ".")
        id = "point";
    else if (c == ",")
        id = "virgule";
    else if (c == "E")
        id = "exposant";
    else if (c == "=")
        id = "calcul";
    else
        id = c;
    Truc.call(this, id, ""+c, ""+c);
    return(this);
}

//class Addition extends Fonction {
derivation(Addition, Fonction);
function Addition() {
    Fonction.call(this, "addition", "+", "+", true, 2);
    return(this);
}
Addition.prototype.evaluer = function(v1, v2) {
    return(v1 + v2);
}

//class Soustraction extends Fonction {
derivation(Soustraction, Fonction);
function Soustraction() {
    Fonction.call(this, "soustraction", "-", "-", true, 2);
    return(this);
}
Soustraction.prototype.evaluer = function(v1, v2) {
    return(v1 - v2);
}

//class Multiplication extends Fonction {
derivation(Multiplication, Fonction);
function Multiplication() {
    Fonction.call(this, "multiplication", "*", "x", true, 2);
    return(this);
}
Multiplication.prototype.evaluer = function(v1, v2) {
    return(v1 * v2);
}

//class Division extends Fonction {
derivation(Division, Fonction);
function Division() {
    Fonction.call(this, "division", "/", "/", true, 2);
    return(this);
}
Division.prototype.evaluer = function(v1, v2) {
    return(v1 / v2);
}

//class Puissance extends Fonction {
derivation(Puissance, Fonction);
function Puissance() {
    Fonction.call(this, "puissance", "^", "^", true, 2);
    return(this);
}
Puissance.prototype.evaluer = function(v1, v2) {
    return(Math.pow(v1, v2));
}

//class Inverse extends Fonction {
derivation(Inverse, Fonction);
function Inverse() {
    Fonction.call(this, "inverse", "inverse", "1/x", false, 1);
    return(this);
}
Inverse.prototype.evaluer = function(v1, v2) {
    return(1/v1);
}

//class PuissanceInverse extends Fonction {
derivation(PuissanceInverse, Fonction);
function PuissanceInverse() {
    Fonction.call(this, "puissanceInverse", "puissanceInverse", "x<sup>1/n<\/sup>", false, 2);
    return(this);
}
PuissanceInverse.prototype.evaluer = function(v1, v2) {
    return(Math.pow(v1, 1/v2));
}

//class Factorielle extends Fonction {
derivation(Factorielle, Fonction);
function Factorielle() {
    Fonction.call(this, "fact", "fact", "x!", false, 1);
    return(this);
}
Factorielle.prototype.evaluer = function(v1, v2) {
    if (Math.floor(v1) == v1) {
        var v = v1;
        while (v1 > 2 && v != Number.POSITIVE_INFINITY) {
            v *= v1-1;
            v1--;
        }
        return(v);
    }
    return(v1);
}

//class M extends Fonction {
derivation(M, Fonction);
function M() {
    Fonction.call(this, "M", "M", "M", false, 1);
    return(this);
}
M.prototype.evaluer = function(v1, v2) {
    memoire = v1;
    return(v1);
}

//class Mplus extends Fonction {
derivation(Mplus, Fonction);
function Mplus() {
    Fonction.call(this, "Mplus", "Mplus", "M+", false, 1);
    return(this);
}
Mplus.prototype.evaluer = function(v1, v2) {
    memoire += v1;
    return(v1);
}

//class RM extends Fonction {
derivation(RM, Fonction);
function RM() {
    Fonction.call(this, "RM", "RM", "RM", false, 0);
    return(this);
}
RM.prototype.evaluer = function(v1, v2) {
    return(memoire);
}

//class Pourcent extends Special {
derivation(Pourcent, Special);
function Pourcent() {
    Special.call(this, "pourcent", "%", "%");
    return(this);
}
Pourcent.prototype.executer = function() {
    try {
        if (edit)
            //valeur1 = (new Double(valeurAffichee)).doubleValue();
            valeur1 = Number(valeurAffichee);
        if (opCourante != null)
            valeur1 = opCourante.evaluer(valeur2, valeur1*valeur2/100);
        else
            valeur1 = valeur1/100;
        valeurAffichee = versString(valeur1);
        expressionCachee = "" + valeur1;
        valeur2 = 0;
        opCourante = null;
        edit = false;
    } catch (ex) {
        alert("pourcent: NumberFormatException: " + ex.description);
    }
}

//class Expressions extends Special {
derivation(Expressions, Special);
function Expressions() {
    Special.call(this, "expressions", "expressions", "expr");
    return(this);
}
Expressions.prototype.executer = function() {
    modeExpressions = !modeExpressions;
}

//class SigneInverse extends Fonction {
derivation(SigneInverse, Fonction);
function SigneInverse() {
    Fonction.call(this, "signeInverse", "signeInverse", "±", false, 1);
    return(this);
}
SigneInverse.prototype.evaluer = function(v1, v2) {
    return(-v1);
}

//class DMS extends Special {
derivation(DMS, Special);
function DMS() {
    Special.call(this, "dms", "dms", "° ' \"");
    return(this);
}
DMS.prototype.executer = function() {
    if (valeurAffichee.indexOf('°') != -1) {
        valeurAffichee = versString(valeur1);
    } else {
        try {
            if (edit)
                //valeur1 = (new Double(valeurAffichee)).doubleValue();
                valeur1 = Number(valeurAffichee);
            var signe = "+";
            if (valeur1 < 0) {
                signe = "-";
                valeur1 = -valeur1;
            }
            var degres = Math.floor(valeur1);
            var dminutes = (valeur1 - degres)*60;
            var minutes = Math.floor(dminutes);
            var secondes = Math.round((dminutes - minutes)*60*100)/100;
            if (secondes >= 60) {
                secondes -= 60;
                minutes += 1;
            }
            if (minutes >= 60) {
                minutes -= 60;
                degres += 1;
            }
            valeurAffichee = signe + degres + "° " + minutes + "' " + secondes + "\"";
            edit = false;
            if ("-" === signe)
                valeur1 = -valeur1;
        } catch (ex) {
            alert("dms: NumberFormatException: " + ex.description);
        }
    }
}

//class DegRad extends Special {
derivation(DegRad, Special);
function DegRad() {
    Special.call(this, "degrad", "degrad", "deg");
    return(this);
}
DegRad.prototype.executer = function() {
    this.titre = degrad;
    if ("rad" === degrad)
        degrad = "deg";
    else
        degrad = "rad";
}

//class Del extends Special {
derivation(Del, Special);
function Del() {
    Special.call(this, "del", "del", "del");
    return(this);
}
Del.prototype.executer = function() {
    if (valeurAffichee.length > 0)
        valeurAffichee = valeurAffichee.substring(0, valeurAffichee.length-1);
}

//class C extends Special {
derivation(C, Special);
function C() {
    Special.call(this, "C", "C", "C");
    return(this);
}
C.prototype.executer = function() {
    valeur1 = 0;
    valeurAffichee = "0";
    edit = false;
}

//class AC extends Special {
derivation(AC, Special);
function AC() {
    Special.call(this, "AC", "AC", "AC");
    return(this);
}
AC.prototype.executer = function() {
    valeur1 = 0;
    valeurAffichee = "0";
    valeur2 = 0;
    opCourante = null;
    edit = false;
    expressionCachee = "";
}

//class Sinus extends Fonction {
derivation(Sinus, Fonction);
function Sinus() {
    Fonction.call(this, "sin", "sin", "sin", false, 1);
    return(this);
}
Sinus.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.sin(v1));
    else
        return(Math.sin(v1/180*Math.PI));
}

//class ArcSinus extends Fonction {
derivation(ArcSinus, Fonction);
function ArcSinus() {
    Fonction.call(this, "asin", "asin", "sin<sup>-1<\/sup>", false, 1);
    return(this);
}
ArcSinus.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.asin(v1));
    else
        return(Math.asin(v1)*180/Math.PI);
}

//class Cosinus extends Fonction {
derivation(Cosinus, Fonction);
function Cosinus() {
    Fonction.call(this, "cos", "cos", "cos", false, 1);
    return(this);
}
Cosinus.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.cos(v1));
    else
        return(Math.cos(v1/180*Math.PI));
}

//class ArcCosinus extends Fonction {
derivation(ArcCosinus, Fonction);
function ArcCosinus() {
    Fonction.call(this, "acos", "acos", "cos<sup>-1<\/sup>", false, 1);
    return(this);
}
ArcCosinus.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.acos(v1));
    else
        return(Math.acos(v1)*180/Math.PI);
}

//class Tangente extends Fonction {
derivation(Tangente, Fonction);
function Tangente() {
    Fonction.call(this, "tan", "tan", "tan", false, 1);
    return(this);
}
Tangente.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.tan(v1));
    else
        return(Math.tan(v1/180*Math.PI));
}

//class ArcTangente extends Fonction {
derivation(ArcTangente, Fonction);
function ArcTangente() {
    Fonction.call(this, "atan", "atan", "tan<sup>-1<\/sup>", false, 1);
    return(this);
}
ArcTangente.prototype.evaluer = function(v1, v2) {
    if (degrad === "rad")
        return(Math.atan(v1));
    else
        return(Math.atan(v1)*180/Math.PI);
}

//class Exponentielle extends Fonction {
derivation(Exponentielle, Fonction);
function Exponentielle() {
    Fonction.call(this, "exp", "exp", "e<sup>x<\/sup>", false, 1);
    return(this);
}
Exponentielle.prototype.evaluer = function(v1, v2) {
    return(Math.exp(v1));
}

//class LogarithmeNeperien extends Fonction {
derivation(LogarithmeNeperien, Fonction);
function LogarithmeNeperien() {
    Fonction.call(this, "ln", "ln", "ln", false, 1);
    return(this);
}
LogarithmeNeperien.prototype.evaluer = function(v1, v2) {
    return(Math.log(v1));
}

//class Exp10 extends Fonction {
derivation(Exp10, Fonction);
function Exp10() {
    Fonction.call(this, "exp10", "exp10", "10<sup>x<\/sup>", false, 1);
    return(this);
}
Exp10.prototype.evaluer = function(v1, v2) {
    return(Math.pow(10,v1));
}

//class LogarithmeDecimal extends Fonction {
derivation(LogarithmeDecimal, Fonction);
function LogarithmeDecimal() {
    Fonction.call(this, "log", "log", "log", false, 1);
    return(this);
}
LogarithmeDecimal.prototype.evaluer = function(v1, v2) {
    return(Math.log(v1)/Math.log(10));
}

//class Carre extends Fonction {
derivation(Carre, Fonction);
function Carre() {
    Fonction.call(this, "sqr", "sqr", "x<sup>2<\/sup>", false, 1);
    return(this);
}
Carre.prototype.evaluer = function(v1, v2) {
    return(v1*v1);
}

//class RacineCarree extends Fonction {
derivation(RacineCarree, Fonction);
function RacineCarree() {
    Fonction.call(this, "sqrt", "sqrt", "√", false, 1);
    return(this);
}
RacineCarree.prototype.evaluer = function(v1, v2) {
    return(Math.sqrt(v1));
}


//class Constante {
function Constante(nom, titre, valeur) {
    this.nom = nom;
    this.titre = titre;
    this.valeur = valeur;
    return(this);
}
Constante.prototype.getNom = function() {
    return(this.nom);
}
Constante.prototype.getTitre = function() {
    return(this.titre);
}
Constante.prototype.getValeur = function() {
    return(this.valeur);
}


//public Double calculer(String s) {
function calculer(s) {
    if ("" === s)
        return(null);
    s = ajParentheses(s);
    param = parser(s);
    d = param.evaluer();
    return(d);
}

//public String ajParentheses(String s) {
function ajParentheses(s) {
    var sops = "^/*-+";
    for (var iops=0; iops<sops.length; iops++) {
        var cops = sops.charAt(iops);
        var indop = s.indexOf(cops);
        if (indop > 0 && "^*/-+Ee".indexOf(s.charAt(indop-1)) != -1)
            indop = -1; // un - ou + après un E ou un opérateur n'est pas un opérateur !
        var nindop = indop;
        var im,ip;
        var cm=' ',cp=' ';
        var pp;
        var ajp;
        while (nindop != -1) {
            ajp = false;
            im = indop - 1;
            if (im >= 0)
                cm = s.charAt(im);
            pp = 0;
            while (im >= 0 && (pp != 0 || cm != '(') &&
                    (pp != 0 || (sops.indexOf(cm) == -1 ||
                    (im > 0 && "^*/-+Ee".indexOf(s.charAt(im-1)) != -1)))) {
                if (cm == ')')
                    pp++;
                else if (cm == '(')
                    pp--;
                im--;
                if (im >= 0)
                    cm = s.charAt(im);
            }
            if (im < 0 || sops.indexOf(cm) != -1)
                ajp = true;
            ip = indop + 1;
            if (ip >= 0 && ip < s.length)
                cp = s.charAt(ip);
            pp = 0;
            while (ip < s.length && (pp != 0 || cp != ')') &&
                    (pp != 0 || (sops.indexOf(cp) == -1 ||
                    (ip > 0 && "^*/-+Ee".indexOf(s.charAt(ip-1)) != -1)))) {
                if (cp == '(')
                    pp++;
                else if (cp == ')')
                    pp--;
                ip++;
                if (ip < s.length)
                    cp = s.charAt(ip);
            }
            if (ip >= s.length || sops.indexOf(cp) != -1)
                ajp = true;
            if (ajp) {
                s = s.substring(0, im+1) + "(" + s.substring(im+1, ip) + ")" +
                    s.substring(ip);
                indop++;
            }
            nindop = s.substring(indop+1).indexOf(cops);
            indop = nindop + indop+1;
        }
    }
    return(s);
}

//public Parametre parser(String s) {
function parser(s) {
    var sops = "^/*-+";
    if (s.length > 0 && s.charAt(0) == '(' && s.charAt(s.length-1) == ')') {
        var retparok = true;
        // on vérifie si on peut retirer les parenthèses qui entourent l'expression
        var pp = 0;
        for (var i=1; i<s.length-1; i++) {
            if (s.charAt(i) == '(')
                pp++;
            else if (s.charAt(i) == ')')
                pp--;
            if (pp < 0) {
                retparok = false;
                break;
            }
        }
        if (retparok)
            s = s.substring(1, s.length-1);
    }
    var indop = -1;
    var pp = 0;
    for (var i=0; i<s.length; i++) {
        if (pp == 0 && sops.indexOf(s.charAt(i)) != -1 &&
                (i==0 || "^*/-+Ee".indexOf(s.charAt(i-1)) == -1)) {
            indop = i;
            break;
        } else if (s.charAt(i) == '(')
            pp++;
        else if (s.charAt(i) == ')')
            pp--;
    }
    if (indop == -1) {
        var nb = true;
        for (var i=0; i<s.length; i++)
            if ("0123456789.,Ee-+ ".indexOf(s.charAt(i)) == -1) {
                nb = false;
                break;
            }
        if (nb) {
            var d;
            try {
                //d = (new Double(s)).doubleValue();
                d = Number(s);
            } catch (ex) {
                d = 0;
            }
            return(new Nombre(d));
        } else {
            var indf = s.indexOf('(');
            if (indf != -1 && s.charAt(s.length-1) == ')') {
                var fct = obtenirFonction(s.substring(0,indf));
                var indv = 0;
                var p1 = null;
                var p2 = null;
                s = s.substring(indf+1, s.length-1);
                indv = s.indexOf(',');
                if (indv != -1) {
                    p1 = parser(s.substring(0,indv).replace(/^\s+|\s+$/g, ""));
                    s = s.substring(indv+1);
                    p2 = parser(s.replace(/^\s+|\s+$/g, ""));
                } else
                    p1 = parser(s.replace(/^\s+|\s+$/g, ""));
                return(new Expression(fct, p1, p2));
            } else
                return(new Variable(s));
        }
    } else {
        var op = obtenirFonction(""+s.charAt(indop));
        var s1 = s.substring(0,indop).replace(/^\s+|\s+$/g, "");
        if ("" === s1)
            s1 = "0"; // -(2+3) = 0 - (2+3)
        var p1 = parser(s1);
        var s2 = s.substring(indop+1).replace(/^\s+|\s+$/g, "");
        var p2 = parser(s2);
        return(new Expression(op, p1, p2));
    }
}

/*
interface Parametre {
    public abstract Double evaluer();
}
*/

//class Expression implements Parametre {
function Expression(fct, p1, p2) {
    this.fct = fct;
    this.p1 = p1;
    this.p2 = p2;
    return(this);
}
Expression.prototype.evaluer = function() {
    if (this.fct == null)
        return(null);
    else if (this.p2 == null) {
        var d1 = this.p1.evaluer();
        if (d1 == null)
            return(null);
        //return(new Double(fct.evaluer(d1.doubleValue(), 0)));
        return(this.fct.evaluer(d1, 0));
    } else {
        var d1 = this.p1.evaluer();
        var d2 = this.p2.evaluer();
        if (d1 == null || d2 == null)
            return(null);
        //return(new Double(fct.evaluer(d1.doubleValue(), d2.doubleValue())));
        return(this.fct.evaluer(d1, d2));
    }
}

//class Variable implements Parametre {
function Variable(nom) {
    this.nom = nom;
    return(this);
}
Variable.prototype.evaluer = function() {
    var v;
    var cst = chercherConstante(this.nom);
    if (cst != null)
        //return(new Double(cst.getValeur()));
        return(cst.getValeur());
    return(null);
}

//class Nombre implements Parametre {
function Nombre(valeur) {
    this.valeur = valeur;
    return(this);
}
Nombre.prototype.evaluer = function() {
    //return(new Double(valeur));
    return(this.valeur);
}


function clicBouton(event) {
    var el1 = (event.target || event.srcElement);
    if (el1 == null)
        return;
    var el2 = parentAvecId(el1);
    if (el2 == null)
        return;
    operation(idVersTruc[el2.id]);
}

function parentAvecId(element) {
    var parent = element;
    while (parent != null && !parent.id && parent != document.body)
        parent = parent.parentNode;
    if (parent == document.body)
        return(null);
    return(parent);
}

function choixConstante(numero) {
    if (numero === 0)
        return;
    var cst = constantes[numero - 1];
    if (modeExpressions) {
        if (edit)
            valeurAffichee += cst.getNom();
        else {
            valeurAffichee = cst.getNom();
            edit = true;
        }
    } else {
        if (opCourante == null)
            expressionCachee = "";
        valeur1 = cst.getValeur();
        valeurAffichee = versString(valeur1);
        edit = true;
    }
    var selectConstantes = document.getElementById('constantes');
    selectConstantes.selectedIndex = 0;
    affichage();
}

function keypress(event) {
    if (event.ctrlKey || event.metaKey)
        return;
    var code;
    if (event.which == null)
        code = event.keyCode;    // old IE
    else if (event.which != 0 && event.charCode != 0)
        code = event.which;	  // All others
    else {
        if (event.keyCode == 13 || event.keyCode == 8)
            code = event.keyCode;
        else
            return; // special key
    }
    var c;
    if (code == 13)
        c = "=";
    else if (code == 8)
        c = "del";
    else {
        c = String.fromCharCode(code);
        if (c == ",")
            c = ".";
        else if (c == "!")
            c = "fact";
    }
    var truc = chercherTruc("" + c);
    if (truc != null) {
        if (event.preventDefault)
            event.preventDefault();
        else
            event.returnValue = false;
        operation(truc);
        truc.activation(true);
    }
}

function keyup(event) {
    for (var i=0; i<trucs.length; i++) {
        if (trucs[i].getActif())
            trucs[i].activation(false);
    }
}


function evenementsBoutons() {
    theTarget = null;
    checkTarget = null;
    if (document.addEventListener) {
        document.addEventListener('mousedown', mousedown, false);
        document.addEventListener('mousemove', mousemove, false);
        document.addEventListener('mouseup', mouseup, false);
    } else {
        document.attachEvent('onmousedown', mousedown);
        document.attachEvent('onmousemove', mousemove);
        document.attachEvent('onmouseup', mouseup);
    }
    if ('ontouchstart' in document) {
        document.addEventListener('touchstart', mousedown, false);
        document.addEventListener('touchmove', mousemove, false);
        document.addEventListener('touchend', mouseup, false);
    }
}

function mousedown(e) {
    // bug IE: lors d'un double clic, le 2ème mousedown n'est pas appelé avec IE<9...
    if (e.type == 'touchstart')
        theTarget = document.elementFromPoint(e.targetTouches[0].clientX, e.targetTouches[0].clientY);
    else
        theTarget = (e.target || e.srcElement);
    if (theTarget != null)
        theTarget = parentAvecId(theTarget);
    if (theTarget != null && idVersTruc[theTarget.id] == null)
        theTarget = null;
    if (theTarget == null)
        return;
    checkTarget = theTarget;
    theTarget.className += ' actif';
    if (e.preventDefault)
        e.preventDefault();
    else
        e.returnValue = false;
    return(false);
}

function mousemove(e) {
    if (theTarget == null)
        return;
    if (e.type == 'touchmove')
        checkTarget = document.elementFromPoint(e.targetTouches[0].clientX, e.targetTouches[0].clientY);
    else
        checkTarget = (e.target || e.srcElement);
    if (checkTarget != null)
        checkTarget = parentAvecId(checkTarget);
    if (e.preventDefault)
        e.preventDefault();
    else
        e.returnValue = false;
    return(false);
}

function mouseup(e) {
    if (theTarget == null)
        return;
    theTarget.className = theTarget.className.replace(/ ?actif/gi, '');
    if (checkTarget == theTarget)
        clicBouton(e);
    theTarget = null;
    if (e.preventDefault)
        e.preventDefault();
    else
        e.returnValue = false;
    return(false);
}
